@@ -1,6 +1,6 @@
 #| repl.jl -- rep input loop
 
-   $Id: repl.jl,v 1.46 2001/10/20 06:02:37 jsh Exp $
+   $Id: repl.jl,v 1.49 2003/09/04 05:58:32 jsh Exp $
 
    Copyright (C) 2000 John Harper <john@dcs.warwick.ac.uk>
 
@@ -27,14 +27,16 @@
 	    make-repl
 	    repl-struct
 	    repl-pending
+	    repl-eval
 	    repl-iterate
-	    repl-completions)
+	    repl-completions
+	    define-repl-command)
 
     (open rep
 	  rep.structures
 	  rep.system
 	  rep.regexp
-	  rep.io.readline)
+	  rep.io.files)
 
   (define current-repl (make-fluid))
 
@@ -88,10 +90,18 @@
 	     (default-error-handler (car data) (cdr data))
 	     t))))))
 
+  (define (do-readline prompt completer)
+    (if (file-ttyp standard-input)
+	(progn
+	  (require 'rep.io.readline)
+	  (readline prompt completer))
+      (write standard-output prompt)
+      (read-line standard-input)))
+
   (define (repl #!optional initial-structure)
     ;; returns t if repl should run again
     (define (run-repl)
-      (let ((input (readline
+      (let ((input (do-readline
 		    (format nil (if (repl-pending (fluid current-repl))
 				    "" "%s> ")
 			    (repl-struct (fluid current-repl)))
@@ -108,7 +118,7 @@
       (let loop ()
 	(when (call-with-exception-handler run-repl interrupt-handler)
 	  (loop)))))
-
+  
   (define (print-list data #!optional map)
     (unless map (setq map identity))
     (let* ((count (length data))
@@ -298,7 +308,17 @@
 		  (intern-structure
 		   (repl-struct (fluid current-repl)))))))
 
-  (define-repl-command 'collect garbage-collect)
+  (define-repl-command
+   'collect
+   (lambda ()
+     (let ((stats (garbage-collect t)))
+       (format standard-output "Used %d/%d cons, %d/%d tuples, %d strings, %d vector slots, %d/%d closures\n"
+	       (car (nth 0 stats)) (+ (car (nth 0 stats)) (cdr (nth 0 stats)))
+	       (car (nth 1 stats)) (+ (car (nth 1 stats)) (cdr (nth 1 stats)))
+	       (car (nth 2 stats))
+	       (nth 3 stats)
+	       (car (nth 4 stats)) (+ (car (nth 4 stats))
+				      (cdr (nth 4 stats)))))))
 
   (define-repl-command
    'disassemble
