@@ -1,6 +1,6 @@
 /* lisp.c -- Core of the Lisp, reading and evaluating...
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: lisp.c,v 1.139 2001/10/20 06:00:39 jsh Exp $
+   $Id: lisp.c,v 1.141 2002/10/05 18:55:22 jsh Exp $
 
    This file is part of Jade.
 
@@ -1590,7 +1590,19 @@ again:
     case rep_SubrN:
 	if (closure)
 	    rep_USE_FUNARG(closure);
-	result = rep_SUBRNFUN(fun)(arglist);
+	if (!rep_SUBR_VEC_P (fun))
+	    result = rep_SUBRNFUN(fun)(arglist);
+	else
+	{
+	    int length;
+	    repv *vec;
+
+	    length = rep_list_length (arglist);
+	    vec = alloca (length * sizeof (repv));
+	    copy_to_vector (arglist, length, vec);
+
+	    result = rep_SUBRVFUN (fun) (length, vec);
+	}
 	break;
 
     case rep_Subr0:
@@ -1770,7 +1782,7 @@ apply FUNCTION ARGS... ARG-LIST
 
 Calls FUNCTION passing all of ARGS to it as well as all elements in ARG-LIST.
 ie,
-  (apply '+ 1 2 3 '(4 5 6))
+  (apply + 1 2 3 '(4 5 6))
    => 21
 ::end:: */
 {
@@ -1889,7 +1901,12 @@ eval(repv obj, repv tail_posn)
 			if (!rep_CONSP (args))
 			    ret = rep_signal_missing_arg (1);
 			else
-			    rep_CDR (args) = Flist_star (rep_CDR (args));
+			{
+			    int len = rep_list_length (rep_CDR (args));
+			    repv *vec = alloca (len * sizeof (repv));
+			    copy_to_vector (rep_CDR (args), len, vec);
+			    rep_CDR (args) = Flist_star (len, vec);
+			}
 		    }
 		    else
 			args = Fcons (funcobj, args);
@@ -2367,6 +2384,18 @@ rep_copy_list(repv list)
     return result;
 }
 
+/* FIXME: required by sawfish; remove at some point */
+repv Fnconc (repv args)
+{
+    int len;
+    repv *vec;
+
+    len = rep_list_length (args);
+    vec = alloca (len * sizeof (repv));
+    copy_to_vector (args, len, vec);
+    return Fnconc_ (len, vec);
+}
+
 /* Used to assign a list of argument values into separate variables.
    Note that optional args without values _are not_ initialized to nil,
    the caller of this function should do that.. */
