@@ -1,6 +1,6 @@
 /* lispcmds.c -- Lots of standard Lisp functions
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: lispcmds.c,v 1.109 2001/08/08 06:15:32 jsh Exp $
+   $Id: lispcmds.c,v 1.112 2002/10/07 07:42:09 jsh Exp $
 
    This file is part of Jade.
 
@@ -143,29 +143,25 @@ is nil.
     return(Qnil);
 }
 
-DEFUN("list", Flist, Slist, (repv args), rep_SubrN) /*
+DEFUN("list", Flist, Slist, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#list::
 list ARGS...
 
 Returns a new list with elements ARGS...
 ::end:: */
 {
-    repv res = Qnil;
-    repv *ptr = &res;
-    while(rep_CONSP(args))
+    repv lst = Qnil;
+    int i;
+
+    for (i = argc - 1; i >= 0; i--)
     {
-	if(!(*ptr = Fcons(rep_CAR(args), Qnil)))
-	    return rep_NULL;
-	ptr = &rep_CDR(*ptr);
-	args = rep_CDR(args);
-	rep_TEST_INT;
-	if(rep_INTERRUPTP)
-	    return rep_NULL;
+	lst = Fcons (argv[i], lst);
     }
-    return res;
+
+    return lst;
 }
 
-DEFUN("list*", Flist_star, Slist_star, (repv args), rep_SubrN) /*
+DEFUN("list*", Flist_star, Slist_star, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#list*::
 list* ARG1 ARG2 ... ARGN
 
@@ -173,24 +169,19 @@ Returns a new list (ARG1 ARG2 ... ARGN-1 . ARGN). That is, the same as from
 `list' but the last argument is dotted to the last but one argument.
 ::end:: */
 {
-    repv res = Qnil;
-    repv *ptr = &res;
-    while(rep_CONSP(args))
+    repv lst;
+    int i;
+
+    if (argc == 0)
+	return Qnil;
+
+    lst = argv[argc - 1];
+    for (i = argc - 2; i >= 0; i--)
     {
-	if(rep_CONSP(rep_CDR(args)))
-	{
-	    if(!(*ptr = Fcons(rep_CAR(args), Qnil)))
-		return rep_NULL;
-	}
-	else
-	    *ptr = rep_CAR(args);
-	ptr = &rep_CDR(*ptr);
-	args = rep_CDR(args);
-	rep_TEST_INT;
-	if(rep_INTERRUPTP)
-	    return rep_NULL;
+	lst = Fcons (argv[i], lst);
     }
-    return res;
+
+    return lst;
 }
 
 DEFUN("make-list", Fmake_list, Smake_list, (repv len, repv init), rep_Subr2) /*
@@ -211,7 +202,7 @@ INITIAL-repv, or nil.
     return(list);
 }
 
-DEFUN("append", Fappend, Sappend, (repv args), rep_SubrN) /*
+DEFUN("append", Fappend, Sappend, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#append::
 append LISTS...
 
@@ -219,35 +210,35 @@ Non-destructively concatenates each of it's argument LISTS... into one
 new list which is returned.
 ::end:: */
 {
-    int i = 1;
-    repv res = Qnil;
-    repv *resend = &res;
-    while(rep_CONSP(args))
+    int i;
+    repv res = Qnil, *res_end = &res;
+
+    for (i = 0; i < argc; i++)
     {
-	if(!rep_LISTP(rep_CAR(args)) && rep_CDR (args) != Qnil)
-	    return rep_signal_arg_error(rep_CAR(args), i);
-	if(rep_CONSP(rep_CAR(args)) && rep_CONSP(rep_CDR(args)))
+	if (i != argc - 1)
 	{
-	    /* Only make a new copy if there's another list after this
-	       one. */
-	    *resend = rep_copy_list(rep_CAR(args));
+	    if (!rep_LISTP(argv[i]))
+		return rep_signal_arg_error (argv[i], i + 1);
+
+	    /* Only make a new copy if there's another list after this one. */
+	    *res_end = rep_copy_list (argv[i]);
 	}
 	else
-	    *resend = rep_CAR(args);	/* Use the old object */
-	while(rep_CONSP(*resend))
+	    *res_end = argv[i];
+
+	while (rep_CONSP (*res_end))
 	{
 	    rep_TEST_INT;
-	    if(rep_INTERRUPTP)
+	    if (rep_INTERRUPTP)
 		return rep_NULL;
-	    resend = &(rep_CDR(*resend));
+	    res_end = rep_CDRLOC (*res_end);
 	}
-	args = rep_CDR(args);
-	i++;
     }
-    return(res);
+
+    return res;
 }
 
-DEFUN("nconc", Fnconc, Snconc, (repv args), rep_SubrN) /*
+DEFUN("nconc", Fnconc_, Snconc, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#nconc::
 nconc LISTS...
 
@@ -256,33 +247,32 @@ list. Every LIST but the last is modified so that it's last cdr points
 to the beginning of the next list. Returns the new list.
 ::end:: */
 {
-    int i = 1;
-    repv res = Qnil;
-    repv *resend = &res;
-    while(rep_CONSP(args))
+    int i;
+    repv res = Qnil, *res_end = &res;
+
+    for (i = 0; i < argc; i++)
     {
-	repv tmp = rep_CAR(args);
-	if(!rep_LISTP(tmp) && rep_CDR (args) != Qnil)
-	    return rep_signal_arg_error(tmp, i);
-	if(rep_CONSP(tmp))
+	if (i != argc - 1)
 	{
-	    if(!rep_CONS_WRITABLE_P(tmp))
-		return Fsignal(Qsetting_constant, rep_LIST_1(tmp));
-	    *resend = tmp;
-	    while(rep_CONSP(rep_CDR(tmp)))
-	    {
-		rep_TEST_INT;
-		if(rep_INTERRUPTP)
-		    return rep_NULL;
-		tmp = rep_CDR(tmp);
-	    }
-	    resend = &rep_CDR(tmp);
+	    if (!rep_LISTP (argv[i]))
+		return rep_signal_arg_error (argv[i], i + 1);
+
+	    if (!rep_CONS_WRITABLE_P (argv[i]))
+		return Fsignal (Qsetting_constant, rep_LIST_1 (argv[i]));
+	}
+
+	*res_end = argv[i];
+
+	while (rep_CONSP (*res_end))
+	{
+	    rep_TEST_INT;
+	    if (rep_INTERRUPTP)
+		return rep_NULL;
+	    res_end = rep_CDRLOC (*res_end);
 	}
-	else
-	    *resend = tmp;
-	args = rep_CDR(args);
     }
-    return(res);
+
+    return res;
 }
 
 DEFUN("rplaca", Frplaca, Srplaca, (repv cons, repv car), rep_Subr2) /*
@@ -815,28 +805,21 @@ applied to that element, ie,
     return list;
 }
 
-DEFUN("vector", Fvector, Svector, (repv args), rep_SubrN) /*
+DEFUN("vector", Fvector, Svector, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#vector::
 vector ARGS...
 
 Returns a new vector with ARGS... as its elements.
 ::end:: */
 {
-    repv res = rep_make_vector(rep_list_length(args));
-    if(res)
+    repv vec = rep_make_vector (argc);
+
+    if(vec != rep_NULL)
     {
-	int i = 0;
-	while(rep_CONSP(args))
-	{
-	    rep_VECTI(res, i) = rep_CAR(args);
-	    args = rep_CDR(args);
-	    i++;
-	    rep_TEST_INT;
-	    if(rep_INTERRUPTP)
-		return(rep_NULL);
-	}
+	memcpy (rep_VECT (vec)->array, argv, argc * sizeof (repv));
     }
-    return(res);
+
+    return vec;
 }
 
 DEFUN("make-vector", Fmake_vector, Smake_vector, (repv size, repv init), rep_Subr2) /*
@@ -985,27 +968,7 @@ All indices start at zero.
 	return(rep_string_dupn(rep_STR(string) + rep_INT(start), slen - rep_INT(start)));
 }
 
-static inline int
-extend_concat(u_char **buf, int *bufLen, int i, int addLen)
-{
-    u_char *newbuf;
-    int newbuflen;
-    if((i + addLen) < *bufLen)
-	return(rep_TRUE);
-    newbuflen = (i + addLen) * 2;
-    newbuf = rep_alloc(newbuflen);
-    if(newbuf)
-    {
-	memcpy(newbuf, *buf, i);
-	rep_free(*buf);
-	*buf = newbuf;
-	*bufLen = newbuflen;
-	return(rep_TRUE);
-    }
-    rep_mem_error();
-    return(rep_FALSE);
-}
-DEFUN("concat", Fconcat, Sconcat, (repv args), rep_SubrN) /*
+DEFUN("concat", Fconcat, Sconcat, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#concat::
 concat ARGS...
 
@@ -1013,82 +976,103 @@ Concatenates all ARGS... into a single string, each argument can be a string,
 a character or a list or vector of characters.
 ::end:: */
 {
-    int buflen = 128;
-    u_char *buf = rep_alloc(buflen);
-    if(buf)
+    u_int length;
+    repv elt, string;
+    u_char *ptr;
+    int i;
+
+    /* Pass 1. calculate the length of the new string. */
+
+    length = 0;
+    for (i = 0; i < argc; i++)
     {
-	repv res = rep_NULL;
-	int i = 0;
-	int argnum = 1;
-	while(rep_CONSP(args))
+	elt = argv[i];
+
+	if (rep_INTP (elt))
 	{
-	    repv arg = rep_CAR(args);
-	    if (arg != Qnil)
+	    length++;
+	}
+	else if (rep_CONSP (elt))
+	{
+	    length += rep_list_length (elt);
+	}
+	else
+	{
+	    switch (rep_CELL8_TYPE (elt))
 	    {
-		switch(rep_TYPE(arg))
-		{
-		    int slen, j;
+	    case rep_String:
+		length += rep_STRING_LEN (elt);
+		break;
 
-		case rep_String:
-		    slen = rep_STRING_LEN(arg);
-		    if(!extend_concat(&buf, &buflen, i, slen))
-			goto error;
-		    memcpy(buf + i, rep_STR(arg), slen);
-		    i += slen;
-		    break;
-
-		case rep_Int:
-		    if(!extend_concat(&buf, &buflen, i, 1))
-			goto error;
-		    buf[i++] = rep_INT(arg);
-		    break;
-
-		case rep_Symbol:
-		case rep_Cons:
-		    while(rep_CONSP(arg))
-		    {
-			repv ch = rep_CAR(arg);
-			if(rep_INTP(ch))
-			{
-			    if(!extend_concat(&buf, &buflen, i, 1))
-				goto error;
-			    buf[i++] = rep_INT(ch);
-			}
-			arg = rep_CDR(arg);
-			rep_TEST_INT;
-			if(rep_INTERRUPTP)
-			    goto error;
-		    }
-		    break;
-
-		case rep_Vector:
-		    {
-			int len = rep_VECT_LEN(arg);
-			for(j = 0; j < len; j++)
-			{
-			    if(rep_INTP(rep_VECTI(arg, j)))
-			    {
-				if(!extend_concat(&buf, &buflen, i, 1))
-				    goto error;
-				buf[i++] = rep_INT(rep_VECTI(arg, j));
-			    }
-			}
-			break;
-		    }
-		default:
-		    res = rep_signal_arg_error(arg, argnum);
-		    goto error;
-		}
+	    case rep_Vector:
+		length += rep_VECT_LEN (elt);
+		break;
 	    }
-	    args = rep_CDR(args);
-	    argnum++;
 	}
-	res = rep_string_dupn(buf, i);
-error:
-	rep_free(buf);
-	return(res);
     }
-    return(rep_NULL);
+
+    if (length == 0)
+	return rep_null_string ();
+
+    /* Allocate the string. */
+
+    string = rep_make_string (length + 1);
+    ptr = rep_STR (string);
+
+    /* Pass 2: copy in the data */
+
+    for (i = 0; i < argc; i++)
+    {
+	elt = argv[i];
+
+	if (rep_INTP (elt))
+	{
+	    *ptr++ = rep_INT (elt);
+	}
+	else if (rep_CONSP (elt))
+	{
+	    repv tem = elt, c;
+
+	    while (rep_CONSP (tem))
+	    {
+		c = rep_CAR (tem);
+
+		if (rep_INTP (c))
+		    *ptr++ = rep_INT (c);
+
+		tem = rep_CDR (tem);
+	    }
+	}
+	else
+	{
+	    switch (rep_CELL8_TYPE (elt))
+	    {
+		int i;
+		repv c;
+
+	    case rep_String:
+		memcpy (ptr, rep_STR (elt), rep_STRING_LEN (elt));
+		ptr += rep_STRING_LEN (elt);
+		break;
+
+	    case rep_Vector:
+		for (i = 0; i < rep_VECT_LEN (elt); i++)
+		{
+		    c = rep_VECTI (elt, i);
+		    if (rep_INTP (c))
+			*ptr++ = rep_INT (c);
+		}
+		break;
+	    }
+	}
+    }
+
+    if (rep_STRING_LEN (string) != (ptr - rep_STR (string)))
+	rep_set_string_len (string, ptr - rep_STR (string));
+
+    *ptr++ = '\0';
+
+    return string;
 }
 
 DEFUN("length", Flength, Slength, (repv sequence), rep_Subr1) /*
@@ -1655,35 +1639,23 @@ Returns t if STRING1 is `less' than STRING2, ignoring case.
     return *s2 ? Qt : Qnil;
 }
 
-
 #define APPLY_COMPARISON(op)				\
-    if(rep_CONSP(args) && rep_CONSP(rep_CDR(args)))	\
+    int i, sign;					\
+    if (argc < 2)					\
+	return rep_signal_missing_arg (argc + 1);	\
+    for (i = 1; i < argc; i++)				\
     {							\
-	repv pred = rep_CAR(args);			\
-	int i = 2;					\
-	args = rep_CDR(args);				\
-	while(rep_CONSP(args))				\
-	{						\
-	    int sign;					\
-	    repv tem = rep_CAR (args);			\
-	    if (rep_NUMBERP (pred) || rep_NUMBERP (tem)) \
-		sign = rep_compare_numbers (pred, tem);	\
-	    else					\
-		sign = rep_value_cmp(pred, tem);	\
-	    if(!(sign op 0)) 				\
-		return Qnil;				\
-	    pred = tem;					\
-	    args = rep_CDR(args);			\
-	    i++;					\
-	    rep_TEST_INT;				\
-	    if(rep_INTERRUPTP)				\
-		return(rep_NULL);			\
-	}						\
-	return Qt;					\
+	repv a = argv[i-1], b = argv[i];		\
+	if (rep_NUMBERP (a) || rep_NUMBERP (b))		\
+	    sign = rep_compare_numbers (a, b);		\
+	else						\
+	    sign = rep_value_cmp (a, b);		\
+	if (!(sign op 0))				\
+	    return Qnil;				\
     }							\
-    return rep_signal_missing_arg(rep_CONSP(args) ? 2 : 1);
+    return Qt;
 
-DEFUN("=", Fnum_eq, Snum_eq, (repv args), rep_SubrN) /*
+DEFUN("=", Fnum_eq, Snum_eq, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#=::
 = ARG1 ARG2 [ARG3 ...]
 
@@ -1695,7 +1667,7 @@ ignored.)
     APPLY_COMPARISON(==)
 }
 
-DEFUN("/=", Fnum_noteq, Snum_noteq, (repv args), rep_SubrN) /*
+DEFUN("/=", Fnum_noteq, Snum_noteq, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#:/=::
 /= ARG1 ARG2 ...
 
@@ -1704,14 +1676,11 @@ Returns t if each value is different from every other value. (Using
 ignored.)
 ::end:: */
 {
-    repv ret = Fnum_eq (args);
-    if (ret)
-	return ret == Qnil ? Qt : Qnil;
-    else
-	return rep_NULL;
+    repv ret = Fnum_eq (argc, argv);
+    return !ret ? rep_NULL : ret == Qnil ? Qt : Qnil;
 }
 
-DEFUN(">", Fgtthan, Sgtthan, (repv args), rep_SubrN) /*
+DEFUN(">", Fgtthan, Sgtthan, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#>::
 > ARG1 ARG2 [ARG3 ...]
 
@@ -1723,7 +1692,7 @@ strings, positions, marks, etc as well.
     APPLY_COMPARISON(>)
 }
 
-DEFUN(">=", Fgethan, Sgethan, (repv args), rep_SubrN) /*
+DEFUN(">=", Fgethan, Sgethan, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#>=::
 >= ARG1 ARG2 [ARG3 ...]
 
@@ -1734,7 +1703,7 @@ isn't limited to numbers, it can do strings, positions, marks, etc as well.
     APPLY_COMPARISON(>=)
 }
 
-DEFUN("<", Fltthan, Sltthan, (repv args), rep_SubrN) /*
+DEFUN("<", Fltthan, Sltthan, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#<::
 < ARG1 ARG2 [ARG3 ...]
 
@@ -1745,7 +1714,7 @@ numbers, it can do strings, positions, marks, etc as well.
     APPLY_COMPARISON(<)
 }
 
-DEFUN("<=", Flethan, Slethan, (repv args), rep_SubrN) /*
+DEFUN("<=", Flethan, Slethan, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.data#<=::
 <= ARG1 ARG2 [ARG3 ...]
 
@@ -2035,8 +2004,10 @@ DEFSTRING(jlc, ".jlc");
 static void
 add_path (const char *env, repv var)
 {
-    repv list = Qnil;
-    char *ptr = getenv (env);
+    repv list = Qnil, vec[2];
+    char *ptr;
+
+    ptr = getenv (env);
     while (ptr != 0 && *ptr != 0)
     {
 	char *end = strchr (ptr, ':');
@@ -2044,8 +2015,10 @@ add_path (const char *env, repv var)
 		      : rep_string_dup (ptr), list);
 	ptr = end ? end + 1 : 0;
     }
-    Fset (var, Fnconc (rep_list_2 (Fnreverse (list),
-				   Fsymbol_value (var, Qt))));
+
+    vec[0] = Fnreverse (list);
+    vec[1] = Fsymbol_value (var, Qt);
+    Fset (var, Fnconc_ (2, vec));
 }
 
 void
