@@ -1,6 +1,6 @@
 /* lispmach.h -- Interpreter for compiled Lisp forms
 
-   $Id: lispmach.h,v 1.8 2002/10/05 18:55:22 jsh Exp $
+   $Id: lispmach.h,v 1.9 2003/09/04 05:58:56 jsh Exp $
 
    Copyright (C) 1993, 1994, 2000 John Harper <john@dcs.warwick.ac.uk>
 
@@ -455,6 +455,11 @@ list_ref (repv list, int elt)
 #define SP_REG asm("27")
 #define SLOTS_REG asm("28")
 #endif
+#if defined (__ppc__) || defined (__powerpc__)
+#define PC_REG asm("r26")
+#define SP_REG asm("r27")
+#define SLOTS_REG asm("r28")
+#endif
 #ifdef __hppa__
 #define PC_REG asm("%r18")
 #define SP_REG asm("%r17")
