@@ -1,6 +1,6 @@
 /* numbers.c -- Implement the tower of numeric types
    Copyright (C) 1993, 1994, 2000 John Harper <john@dcs.warwick.ac.uk>
-   $Id: numbers.c,v 1.58 2002/02/19 05:33:53 jsh Exp $
+   $Id: numbers.c,v 1.63 2002/10/07 07:42:09 jsh Exp $
 
    This file is part of librep.
 
@@ -1237,6 +1237,29 @@ rep_number_foldl (repv args, repv (*op)(repv, repv))
 	    : rep_signal_missing_arg (1));
 }
 
+static inline repv
+number_foldv (int argc, repv *argv, repv (*op) (repv, repv))
+{
+    repv sum;
+    int i;
+
+    if (argc < 1)
+	return rep_signal_missing_arg (1);
+    if (!rep_NUMERICP (argv[0]))
+	return rep_signal_arg_error (argv[0], 1);
+
+    sum = argv[0];
+    for (i = 1; i < argc; i++)
+    {
+	if (!rep_NUMERICP (argv[i]))
+	    return rep_signal_arg_error (argv[i], i + 1);
+
+	sum = op (sum, argv[i]);
+    }
+
+    return sum;
+}
+
 repv
 rep_integer_foldl (repv args, repv (*op)(repv, repv))
 {
@@ -1260,6 +1283,29 @@ rep_integer_foldl (repv args, repv (*op)(repv, repv))
 	    : rep_signal_missing_arg (1));
 }
 
+static inline repv
+integer_foldv (int argc, repv *argv, repv (*op) (repv, repv))
+{
+    repv sum;
+    int i;
+
+    if (argc < 1)
+	return rep_signal_missing_arg (1);
+    if (!rep_INTEGERP (argv[0]))
+	return rep_signal_arg_error (argv[0], 1);
+
+    sum = argv[0];
+    for (i = 1; i < argc; i++)
+    {
+	if (!rep_INTEGERP (argv[i]))
+	    return rep_signal_arg_error (argv[i], i + 1);
+
+	sum = op (sum, argv[i]);
+    }
+
+    return sum;
+}
+
 repv
 rep_foldl (repv args, repv (*op)(repv, repv))
 {
@@ -1280,6 +1326,24 @@ rep_foldl (repv args, repv (*op)(repv, repv))
     return rep_signal_missing_arg (1);
 }
 
+static inline repv
+foldv (int argc, repv *argv, repv (*op) (repv, repv))
+{
+    repv sum;
+    int i;
+
+    if (argc < 1)
+	return rep_signal_missing_arg (1);
+
+    sum = argv[0];
+    for (i = 1; i < argc; i++)
+    {
+	sum = op (sum, argv[i]);
+    }
+
+    return sum;
+}
+
 repv
 rep_number_add (repv x, repv y)
 {
@@ -1725,20 +1789,20 @@ rep_integer_gcd (repv x, repv y)
     return out;
 }
 
-DEFUN("+", Fplus, Splus, (repv args), rep_SubrN) /*
+DEFUN("+", Fplus, Splus, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#+::
 + NUMBERS...
 
 Adds all NUMBERS together. If no arguments are given returns 0.
 ::end:: */
 {
-    if (args == Qnil)
+    if (argc == 0)
 	return rep_MAKE_INT (0);
     else
-	return rep_number_foldl (args, rep_number_add);
+	return number_foldv (argc, argv, rep_number_add);
 }
 
-DEFUN("-", Fminus, Sminus, (repv args), rep_SubrN) /*
+DEFUN("-", Fminus, Sminus, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#-::
 - NUMBER [NUMBERS...]
 
@@ -1746,39 +1810,40 @@ Either returns the negation of NUMBER or the value of NUMBER minus
 NUMBERS
 ::end:: */
 {
-    if (args == Qnil)
+    if (argc == 0)
 	return rep_signal_missing_arg (1);
-    else if (!rep_CONSP (rep_CDR (args)))
-	return rep_number_neg (rep_CAR (args));
+    else if (argc == 1)
+	return rep_number_neg (argv[0]);
     else
-	return rep_number_foldl (args, rep_number_sub);
+	return number_foldv (argc, argv, rep_number_sub);
 }
 
-DEFUN("*", Fproduct, Sproduct, (repv args), rep_SubrN) /*
+DEFUN("*", Fproduct, Sproduct, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#*::
 * NUMBERS...
 
 Multiplies all NUMBERS together. If no numbers are given returns 1.
 ::end:: */
 {
-    if (args == Qnil)
+    if (argc == 0)
 	return rep_MAKE_INT (1);
     else
-	return rep_number_foldl (args, rep_number_mul);
+	return number_foldv (argc, argv, rep_number_mul);
 }
 
-DEFUN("/", Fdivide, Sdivide, (repv args), rep_SubrN) /*
+DEFUN("/", Fdivide, Sdivide, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#/::
 / NUMBERS...
 
 Divides NUMBERS (in left-to-right order).
 ::end:: */
 {
-    if (args == Qnil)
+    if (argc == 0)
 	return rep_signal_missing_arg (1);
-    else if (!rep_CONSP (rep_CDR (args)))
-	return rep_number_div (rep_MAKE_INT (1), rep_CAR (args));
-    return rep_number_foldl (args, rep_number_div);
+    else if (argc == 1)
+	return rep_number_div (rep_MAKE_INT (1), argv[0]);
+    else
+	return number_foldv (argc, argv, rep_number_div);
 }
 
 DEFUN("remainder", Fremainder, Sremainder, (repv n1, repv n2), rep_Subr2) /*
@@ -1921,37 +1986,37 @@ Returns the bitwise logical `not' of NUMBER.
     return rep_number_lognot (num);
 }
 
-DEFUN("logior", Flogior, Slogior, (repv args), rep_SubrN) /*
+DEFUN("logior", Flogior, Slogior, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#logior::
 logior NUMBERS...
 
 Returns the bitwise logical `inclusive-or' of its arguments.
 ::end:: */
 {
-    if (args == Qnil)
+    if (argc == 0)
 	return rep_MAKE_INT (0);
     else
-	return rep_number_foldl (args, rep_number_logior);
+	return number_foldv (argc, argv, rep_number_logior);
 }
 
-DEFUN("logxor", Flogxor, Slogxor, (repv args), rep_SubrN) /*
+DEFUN("logxor", Flogxor, Slogxor, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#logxor::
 logxor NUMBERS...
 
 Returns the bitwise logical `exclusive-or' of its arguments.
 ::end:: */
 {
-    return rep_number_foldl (args, rep_number_logxor);
+    return number_foldv (argc, argv, rep_number_logxor);
 }
 
-DEFUN("logand", Flogand, Slogand, (repv args), rep_SubrN) /*
+DEFUN("logand", Flogand, Slogand, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#logand::
 logand NUMBERS...
 
 Returns the bitwise logical `and' of its arguments.
 ::end:: */
 {
-    return rep_number_foldl (args, rep_number_logand);
+    return number_foldv (argc, argv, rep_number_logand);
 }
 
 DEFUN("eql", Feql, Seql, (repv arg1, repv arg2), rep_Subr2) /*
@@ -2492,7 +2557,7 @@ signalled (mathematically should return a complex number).
     return out;
 }
 
-DEFUN("gcd", Fgcd, Sgcd, (repv args), rep_SubrN) /*
+DEFUN("gcd", Fgcd, Sgcd, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#gcd::
 gcd ...
 
@@ -2500,15 +2565,15 @@ Return the greatest common divisor of the integer arguments. The result
 is always non-negative. Returns 0 with arguments.
 ::end:: */
 {
-    if (args == Qnil)
+    if (argc == 0)
 	return rep_MAKE_INT (0);
-    else if (rep_CONSP (args) && rep_CDR (args) == Qnil)
+    else if (argc == 1)
     {
-	rep_DECLARE1 (rep_CAR (args), rep_INTEGERP);
-	return rep_integer_gcd (rep_CAR (args), rep_CAR (args));
+	rep_DECLARE1 (argv[0], rep_INTEGERP);
+	return rep_integer_gcd (argv[0], argv[0]);
     }
     else
-	return rep_integer_foldl (args, rep_integer_gcd);
+	return integer_foldv (argc, argv, rep_integer_gcd);
 }
 
 DEFUN("numberp", Fnumberp, Snumberp, (repv arg), rep_Subr1) /*
@@ -2583,6 +2648,32 @@ Returns an inexact (i.e. floating point) representation of X.
 	return rep_make_float (rep_get_float (arg), rep_TRUE);
 }
 
+static void
+rationalize (repv arg, double *numerator, double *denominator)
+{
+    double x, y;
+    int expt;
+ 
+    /* X/Y always equals the input value. Tactic is to iteratively
+       multiply both X and Y by 2 until X is an integer. We bound
+       the number of iterations to the size of the mantissa
+       by starting with the normalized value... */
+
+    x = frexp (rep_get_float (arg), &expt);
+    y = pow (2.0, -expt);
+
+    while (x - floor (x) > DBL_EPSILON)
+    {
+	x = x * 2.0;
+	y = y * 2.0;
+    }
+
+    if (numerator != NULL)
+	*numerator = x;
+    if (denominator != NULL)
+	*denominator = y;
+}
+
 DEFUN("inexact->exact", Finexact_to_exact,
       Sinexact_to_exact, (repv arg), rep_Subr1) /*
 ::doc:rep.lang.math#inexact->exact::
@@ -2593,46 +2684,37 @@ accuracy.
 ::end:: */
 {
     rep_DECLARE1(arg, rep_NUMERICP);
+
     if (rep_INTP (arg) || !rep_NUMBER_FLOAT_P (arg))
 	return arg;
+
+#ifdef HAVE_GMP
     else
     {
-	double x;
-	repv y;
-	int expt;
+	rep_number_q *q;
+
+	q = make_number (rep_NUMBER_RATIONAL);
+	mpq_init (q->q);
+	mpq_set_d (q->q, rep_get_float (arg));
+
+	return maybe_demote (rep_VAL (q));
+    }
+#else
+    else
+    {
+	double x, y;
 	rep_number_z *z;
 
-	/* X/Y always equals the input value. Tactic is to iteratively
-	   multiply both X and Y by 2 until X is an integer. We bound
-	   the number of iterations to the size of the mantissa
-	   by starting with the normalized value... */
-
-	x = frexp (rep_get_float (arg), &expt);
-	y = Fexpt (rep_MAKE_INT (2), rep_make_long_int (-expt));
-
-	while (x - floor (x) > DBL_EPSILON)
-	{
-	    x = x * 2.0;
-	    y = rep_number_mul (y, rep_MAKE_INT (2));
-	}
-
+	rationalize (arg, &x, &y);
 	z = make_number (rep_NUMBER_BIGNUM);
-#ifdef HAVE_GMP
-	mpz_init_set_d (z->z, x);
-#else
-	if (x >= BIGNUM_MAX)
-	    z->z = BIGNUM_MAX;
-	else if (x <= BIGNUM_MIN)
-	    z->z = BIGNUM_MIN;
-	else
-	    z->z = (rep_long_long) x;
-#endif
+	z->z = x / y;
 
-	return rep_number_div (rep_VAL (z), y);
+	return maybe_demote (rep_VAL (z));
     }
+#endif
 }
 
-DEFUN("numerator", Fnumerator, Snumerator, (repv x), rep_Subr1) /*
+DEFUN("numerator", Fnumerator, Snumerator, (repv arg), rep_Subr1) /*
 ::doc:rep.lang.math#numerator::
 numerator X
 
@@ -2640,30 +2722,28 @@ Return the numerator of rational number X.
 ::end:: */
 {
     rep_bool inexact = rep_FALSE;
-    rep_DECLARE1(x, rep_NUMERICP);
-    if (rep_INTP (x) || rep_NUMBER_BIGNUM_P (x))
-	return x;
-    else if (rep_NUMBER_FLOAT_P (x))
-    {
-	x = Finexact_to_exact (x);
-	inexact = rep_TRUE;
-    }
+    double x;
+
+    rep_DECLARE1(arg, rep_NUMERICP);
+
 #ifdef HAVE_GMP
-    if (rep_NUMBER_RATIONAL_P (x))
+    if (rep_NUMBER_RATIONAL_P (arg))
     {
 	rep_number_z *z = make_number (rep_NUMBER_BIGNUM);
-	mpz_init_set (z->z, mpq_numref (rep_NUMBER(x,q)));
-	if (!inexact)
-	    return maybe_demote (rep_VAL (z));
-	else
-	    return rep_make_float (mpz_get_d (z->z), rep_TRUE);
+	mpz_init_set (z->z, mpq_numref (rep_NUMBER(arg,q)));
+	return maybe_demote (rep_VAL (z));
     }
 #endif
-    else
-	return rep_signal_arg_error (x, 1);
+
+    if (rep_NUMBER_INEXACT_P (arg))
+	inexact = rep_TRUE;
+
+    rationalize (arg, &x, NULL);
+
+    return rep_make_float (x, inexact);
 }
 
-DEFUN("denominator", Fdenominator, Sdenominator, (repv x), rep_Subr1) /*
+DEFUN("denominator", Fdenominator, Sdenominator, (repv arg), rep_Subr1) /*
 ::doc:rep.lang.math#denominator::
 denominator X
 
@@ -2671,30 +2751,28 @@ Return the denominator of rational number X.
 ::end:: */
 {
     rep_bool inexact = rep_FALSE;
-    rep_DECLARE1(x, rep_NUMERICP);
-    if (rep_INTP (x) || rep_NUMBER_BIGNUM_P (x))
-	return rep_MAKE_INT (1);
-    else if (rep_NUMBER_FLOAT_P (x))
-    {
-	x = Finexact_to_exact (x);
-	inexact = rep_TRUE;
-    }
+    double y;
+
+    rep_DECLARE1(arg, rep_NUMERICP);
+
 #ifdef HAVE_GMP
-    if (rep_NUMBER_RATIONAL_P (x))
+    if (rep_NUMBER_RATIONAL_P (arg))
     {
 	rep_number_z *z = make_number (rep_NUMBER_BIGNUM);
-	mpz_init_set (z->z, mpq_denref (rep_NUMBER(x,q)));
-	if (!inexact)
-	    return maybe_demote (rep_VAL (z));
-	else
-	    return rep_make_float (mpz_get_d (z->z), rep_TRUE);
+	mpz_init_set (z->z, mpq_denref (rep_NUMBER(arg,q)));
+	return maybe_demote (rep_VAL (z));
     }
 #endif
-    else
-	return rep_signal_arg_error (x, 1);
+
+    if (rep_NUMBER_INEXACT_P (arg))
+	inexact = rep_TRUE;
+
+    rationalize (arg, NULL, &y);
+
+    return rep_make_float (y, inexact);
 }
 
-DEFUN("max", Fmax, Smax, (repv args), rep_SubrN) /*
+DEFUN("max", Fmax, Smax, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#max::
 max ARGS...
 
@@ -2703,10 +2781,10 @@ arguments. When comparing numbers, any inexact arguments cause the
 result to be inexact.
 ::end:: */
 {
-    return rep_foldl (args, rep_number_max);
+    return foldv (argc, argv, rep_number_max);
 }
 
-DEFUN("min", Fmin, Smin, (repv args), rep_SubrN) /*
+DEFUN("min", Fmin, Smin, (int argc, repv *argv), rep_SubrV) /*
 ::doc:rep.lang.math#min::
 min ARGS...
 
@@ -2715,7 +2793,7 @@ arguments. When comparing numbers, any inexact arguments cause the
 result to be inexact.
 ::end:: */
 {
-    return rep_foldl (args, rep_number_min);
+    return foldv (argc, argv, rep_number_min);
 }
 
 DEFUN("string->number", Fstring_to_number,
@@ -2829,7 +2907,7 @@ in base 10.
 
 /* Random number generation */
 
-#if defined (HAVE_GMP) && defined (HAVE_GMP_RANDINIT)
+#if defined (HAVE_GMP) && defined (HAVE_GMP_RANDINIT) && __GNU_MP__ >= 4
 
 static gmp_randstate_t random_state;
 
