@@ -1,6 +1,6 @@
 /* readline.c -- wrap some readline functions when available
    Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
-   $Id: readline.c,v 1.13 2001/10/20 06:00:40 jsh Exp $
+   $Id: readline.c,v 1.14 2003/10/28 22:29:39 jsh Exp $
 
    This file is part of librep.
 
@@ -22,7 +22,14 @@
 
 #include <config.h>
 #include <rep.h>
-#include <sys/time.h>
+
+#ifdef HAVE_SYS_TIME_H
+# include <sys/time.h>
+#endif
+
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
 
 #ifdef HAVE_LIBREADLINE
 # include <readline/readline.h>
