@@ -1,6 +1,6 @@
 /* rep_lisp.h -- Data structures/objects for Lisp
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: rep_lisp.h,v 1.62 2000/07/15 11:41:05 john Exp $
+   $Id: rep_lisp.h,v 1.72 2002/10/05 18:55:22 jsh Exp $
 
    This file is part of Jade.
 
@@ -390,7 +390,7 @@ typedef struct rep_string_struct {
 /* Define a variable V, containing a static string S. This must be cast
    to a repv via the rep_VAL() macro when using. */
 #define DEFSTRING(v, s)					\
-    rep_ALIGN_CELL(const static rep_string v) = {	\
+    rep_ALIGN_CELL(static const rep_string v) = {	\
 	((sizeof(s) - 1) << rep_STRING_LEN_SHIFT)	\
 	| rep_CELL_STATIC_BIT | rep_String,		\
 	(u_char *)s					\
@@ -411,6 +411,8 @@ typedef struct {
     repv name;
 } rep_symbol;
 
+#define rep_SF_KEYWORD	(1 << (rep_CELL8_TYPE_BITS + 0))
+
 /* Means that the symbol's value may be in some form of local storage,
    if so then that occurrence takes precedence. */
 #define rep_SF_LOCAL 	(1 << (rep_CELL8_TYPE_BITS + 1))
@@ -438,12 +440,19 @@ typedef struct {
 /* Set when the variable has been defvar'd */
 #define rep_SF_DEFVAR	(1 << (rep_CELL8_TYPE_BITS + 7))
 
+#define rep_SF_LITERAL	(1 << (rep_CELL8_TYPE_BITS + 8))
+
 #define rep_SYM(v)		((rep_symbol *)rep_PTR(v))
 #define rep_SYMBOLP(v)		rep_CELL8_TYPEP(v, rep_Symbol)
 
 #define rep_NILP(v)		((v) == Qnil)
 #define rep_LISTP(v)		(rep_NILP(v) || rep_CONSP(v))
 
+#define rep_KEYWORDP(v)		(rep_SYMBOLP(v) \
+				 && (rep_SYM(v)->car & rep_SF_KEYWORD) != 0)
+
+#define rep_SYMBOL_LITERAL_P(v)	((rep_SYM(v)->car & rep_SF_LITERAL) != 0)
+
 
 /* Vectors */
 
@@ -473,27 +482,24 @@ typedef struct rep_vector_struct {
 #define rep_COMPILEDP(v)	rep_CELL8_TYPEP(v, rep_Compiled)
 #define rep_COMPILED(v)		((rep_vector *)rep_PTR(v))
 
-/* First element is lambda list. */
-#define rep_COMPILED_LAMBDA(v)	rep_VECTI(v, 0)
+/* First elt is byte-code string */
+#define rep_COMPILED_CODE(v)	rep_VECTI(v, 0)
 
-/* Second is byte-code string */
-#define rep_COMPILED_CODE(v)	rep_VECTI(v, 1)
+/* Second is constant vector */
+#define rep_COMPILED_CONSTANTS(v) rep_VECTI(v, 1)
 
-/* Third is constant vector */
-#define rep_COMPILED_CONSTANTS(v) rep_VECTI(v, 2)
+/* Third is an (opaque) integer: memory requirements */
+#define rep_COMPILED_STACK(v)	rep_VECTI(v, 2)
 
-/* Fourth is an integer: stack usage */
-#define rep_COMPILED_STACK(v)	rep_VECTI(v, 3)
-
-#define rep_COMPILED_MIN_SLOTS	4
+#define rep_COMPILED_MIN_SLOTS	3
 
 /* Optional fifth element is documentation. */
-#define rep_COMPILED_DOC(v)	((rep_VECT_LEN(v) >= 5) \
-				 ? rep_VECTI(v, 4) : Qnil)
+#define rep_COMPILED_DOC(v)	((rep_VECT_LEN(v) >= 4) \
+				 ? rep_VECTI(v, 3) : Qnil)
 
 /* Optional sixth element is interactive specification. */
-#define rep_COMPILED_INTERACTIVE(v) ((rep_VECT_LEN(v) >= 6) \
-				     ? rep_VECTI(v, 5) : Qnil)
+#define rep_COMPILED_INTERACTIVE(v) ((rep_VECT_LEN(v) >= 5) \
+				     ? rep_VECTI(v, 4) : Qnil)
 
 
 /* Files */
@@ -521,11 +527,15 @@ typedef struct rep_file_struct {
 	repv stream;
     } file;
 
+    /* For input streams */
+    int line_number;
 } rep_file;
 
 /* When this bit is set in flags, the file handle is never fclose()'d,
    i.e. this file points to something like stdin. */
-#define rep_LFF_DONT_CLOSE	(1 << rep_CELL16_TYPE_BITS)
+#define rep_LFF_DONT_CLOSE	(1 << (rep_CELL16_TYPE_BITS + 0))
+#define rep_LFF_BOGUS_LINE_NUMBER (1 << (rep_CELL16_TYPE_BITS + 1))
+#define rep_LFF_SILENT_ERRORS	(1 << (rep_CELL16_TYPE_BITS + 2))
 
 #define rep_FILE(v)		((rep_file *)rep_PTR(v))
 #define rep_FILEP(v)		rep_CELL16_TYPEP(v, rep_file_type)
@@ -550,6 +560,7 @@ typedef struct {
 	repv (*fun3)(repv, repv, repv);
 	repv (*fun4)(repv, repv, repv, repv);
 	repv (*fun5)(repv, repv, repv, repv, repv);
+	repv (*funv)(int, repv *);
     } fun;
     repv name;
     repv int_spec;
@@ -562,6 +573,12 @@ typedef struct {
     repv int_spec;			/* put this in plist? */
 } rep_xsubr;
 
+/* If set in rep_SubrN types, it'll be passed a vector of args,
+   instead of a list */
+#define rep_SUBR_VEC      (1 << (rep_CELL8_TYPE_BITS + 0))
+#define rep_SUBR_VEC_P(v) (rep_SUBR(v)->car & rep_SUBR_VEC)
+#define rep_SubrV         (rep_SubrN | rep_SUBR_VEC)
+
 #define rep_XSUBR(v)	((rep_xsubr *) rep_PTR(v))
 #define rep_SUBR(v)	((rep_subr *) rep_PTR(v))
 #define rep_SUBR0FUN(v)	(rep_SUBR(v)->fun.fun0)
@@ -571,6 +588,7 @@ typedef struct {
 #define rep_SUBR4FUN(v)	(rep_SUBR(v)->fun.fun4)
 #define rep_SUBR5FUN(v)	(rep_SUBR(v)->fun.fun5)
 #define rep_SUBRNFUN(v)	(rep_SUBR(v)->fun.fun1)
+#define rep_SUBRVFUN(v)	(rep_SUBR(v)->fun.funv)
 #define rep_SFFUN(v)	(rep_SUBR(v)->fun.fun2)
 
 
@@ -755,9 +773,9 @@ typedef struct rep_gc_n_roots {
 #define DEFUN(name,fsym,ssym,args,type)					\
     DEFSTRING(rep_CONCAT(ssym, __name), name);				\
     extern repv fsym args;						\
-    rep_ALIGN_CELL(rep_xsubr ssym) = { type, fsym,			\
-					rep_VAL(&rep_CONCAT(ssym, __name)), \
-					rep_NULL };			\
+    rep_ALIGN_CELL(rep_xsubr ssym) = { type, (repv (*)()) fsym,		\
+				       rep_VAL(&rep_CONCAT(ssym, __name)), \
+				       rep_NULL };			\
     repv fsym args
 
 /* Same as above but with an extra arg -- an interactive-spec string. */
@@ -765,9 +783,9 @@ typedef struct rep_gc_n_roots {
     DEFSTRING(rep_CONCAT(ssym, __name), name);				\
     DEFSTRING(rep_CONCAT(ssym, __int), interactive);			\
     extern repv fsym args;						\
-    rep_ALIGN_CELL(rep_xsubr ssym) = { type, fsym,			\
-					rep_VAL(&rep_CONCAT(ssym, __name)), \
-					rep_VAL(&rep_CONCAT(ssym, __int)) };\
+    rep_ALIGN_CELL(rep_xsubr ssym) = { type, (repv (*)()) fsym,		\
+				       rep_VAL(&rep_CONCAT(ssym, __name)), \
+				       rep_VAL(&rep_CONCAT(ssym, __int)) };\
     repv fsym args
 
 /* Add a subroutine */    
@@ -818,6 +836,12 @@ typedef struct rep_gc_n_roots {
 #define rep_DECLARE4(x,t) rep_DECLARE(4,x,t(x))
 #define rep_DECLARE5(x,t) rep_DECLARE(5,x,t(x))
 
+#define rep_DECLARE1_OPT(x,t) rep_DECLARE(1, x, (x) == Qnil || t(x))
+#define rep_DECLARE2_OPT(x,t) rep_DECLARE(2, x, (x) == Qnil || t(x))
+#define rep_DECLARE3_OPT(x,t) rep_DECLARE(3, x, (x) == Qnil || t(x))
+#define rep_DECLARE4_OPT(x,t) rep_DECLARE(4, x, (x) == Qnil || t(x))
+#define rep_DECLARE5_OPT(x,t) rep_DECLARE(5, x, (x) == Qnil || t(x))
+
 
 /* Macros for interrupt handling */
 
@@ -866,6 +890,28 @@ typedef struct rep_gc_n_roots {
    should exit as soon as possible, returning rep_NULL. */
 #define rep_INTERRUPTP (rep_throw_value != rep_NULL)
 
+
+/* End-of-list / false value
+
+   The canonical method of getting '() is to access the `Qnil' variable.
+
+   But we know that that currently points to `rep_eol_datum'. So avoid
+   lots of global variable referencing by hardcoding that value for
+   library-internal code. */
+
+extern repv Qnil;
+
+#ifdef rep_INTERNAL
+  extern rep_tuple rep_eol_datum;
+# ifdef rep_DEFINE_QNIL
+    repv Qnil = rep_VAL (&rep_eol_datum);
+# endif
+  /* OS X has problems with this */
+# ifndef __APPLE__
+#  define Qnil rep_VAL(&rep_eol_datum)
+# endif
+#endif
+
 
 /* Storing timestamps */
 
